// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: ft_allocator.cc,v 1.1 2002/09/29 05:50:47 carzanig Exp $
//
#include <cstdlib>

#include "ft_allocator.h"

struct align_tester {
    char c;
    void *v;
};

static const unsigned char Align = sizeof(align_tester) - sizeof(void*);

static inline size_t round_up_to_alignment(size_t s) {
    return (s + Align - 1) & ~(Align - 1);
}

size_t FTAllocator::size() const {
    return bcount * BSize;
}

void * FTAllocator::allocate(size_t s) {
    size_t new_fp = round_up_to_alignment(free_pos + s);
    if (new_fp > BSize || blist == NULL) {
	block * nb = static_cast<block*>(malloc(BSize));
	if (nb == NULL) 
	    return NULL;

	nb->next = blist;
	blist = nb;
	free_pos = 0;
	++bcount;
	new_fp = round_up_to_alignment(s);
    } 
    void  * res = &(blist->bytes[free_pos]);
    free_pos = new_fp;
    return res;
}

void FTAllocator::clear() {
    block * tmp;
    while (blist != NULL) {
	tmp = blist;
	blist = blist->next;
	free(tmp);
    }
    free_pos = 0;
    bcount = 0;
    blist = NULL;
}
