// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: string_index.cc,v 1.9 2002/09/29 15:02:48 carzanig Exp $
//
#include <iostream>
#include <set>

#include "fwdconf.h"
#include "fwd_table.h"
#include "string_index.h"
#include "ft_allocator.h"

#ifndef HAVE_INLINE
#include "string_index.icc"
#endif

Constraint * StringIndex::add_eq(const string & v, FTAllocator & ftmemory) { 
    TSTNode * np = insert(v.c_str(), ftmemory);
    SICompleteDescr * d;
    if (np->data == NULL) {
	d = new (ftmemory)SICompleteDescr();
	np->data = d;
	d->eq = new (ftmemory)Constraint();
	d->lt = NULL;
	d->gt = NULL;
	d->sf = NULL;
	d->insert_between(np->prev(), np->next());
	return d->eq;
    }
    d = static_cast<SICompleteDescr*>(np->data);
    if (d->eq == NULL) {
	d->eq = new (ftmemory)Constraint();
    }
    return d->eq;
}

Constraint * StringIndex::add_sf(const string & v, FTAllocator & ftmemory) { 
    TSTNode * np = insert(v.c_str(), ftmemory);
    SICompleteDescr * d;
    if (np->data == NULL) {
	d = new (ftmemory)SICompleteDescr();
	np->data = d;
	d->eq = NULL;
	d->lt = NULL;
	d->gt = NULL;
	d->sf = new (ftmemory)Constraint();
	d->insert_between(np->prev(), np->next());
	return d->sf;
    }
    d = static_cast<SICompleteDescr*>(np->data);
    if (d->sf == NULL) {
	d->sf = new (ftmemory)Constraint();
    }
    return d->sf;
}

Constraint * StringIndex::add_lt(const string & v, FTAllocator & ftmemory) { 
    TSTNode * np = insert(v.c_str(), ftmemory);
    SICompleteDescr * d;
    if (np->data == NULL) {
	d = new (ftmemory)SICompleteDescr();
	np->data = d;
	d->eq = NULL;
	d->lt = new (ftmemory)Constraint();
	d->gt = NULL;
	d->sf = NULL;
	d->insert_between(np->prev(), np->next());
	return 	d->lt;
    }
    d = static_cast<SICompleteDescr*>(np->data);
    if (d->lt == NULL) {
	d->lt = new (ftmemory)Constraint();
	d->insert_between(np->prev(), np->next());
    }
    return d->lt;
}

Constraint * StringIndex::add_gt(const string & v, FTAllocator & ftmemory) { 
    TSTNode * np = insert(v.c_str(), ftmemory);
    SICompleteDescr * d;
    if (np->data == NULL) {
	d = new (ftmemory)SICompleteDescr();
	np->data = d;
	d->eq = NULL;
	d->lt = NULL;
	d->gt = new (ftmemory)Constraint();
	d->sf = NULL;
	d->insert_between(np->prev(), np->next());
	return 	d->gt;
    }
    d = static_cast<SICompleteDescr*>(np->data);
    if (d->gt == NULL) {
	d->gt = new (ftmemory)Constraint();
	d->insert_between(np->prev(), np->next());
    }
    return d->gt;
}

Constraint * StringIndex::add_pf(const string & v, FTAllocator & ftmemory) { 
    TSTNode * np = insert_partial(v.c_str(), ftmemory);
    SIPartialDescr * d;
    if (np->data == NULL) {
	d = new (ftmemory)SIPartialDescr();
	np->data = d;
	d->ss = NULL;
	d->pf = new (ftmemory)Constraint();
	return d->pf;
    }
    d = static_cast<SIPartialDescr*>(np->data);
    if (d->pf == NULL) {
	d->pf = new (ftmemory)Constraint();
    }
    return d->pf;
}

Constraint * StringIndex::add_ss(const string & v, FTAllocator & ftmemory) { 
    TSTNode * np = insert_partial(v.c_str(), ftmemory);
    SIPartialDescr * d;
    if (np->data == NULL) {
	d = new (ftmemory)SIPartialDescr();
	np->data = d;
	d->pf = NULL;
	d->ss = new (ftmemory)Constraint();
	return d->ss;
    }
    d = static_cast<SIPartialDescr*>(np->data);
    if (d->ss == NULL) {
	d->ss = new (ftmemory)Constraint();
    }
    return d->ss;
}

void SICompleteDescr::insert_between(TSTNode * prev, TSTNode * next) {
    if (prev == NULL) {
	next_gt = NULL;
    } else {
	SICompleteDescr * prevd;
	prevd = static_cast<SICompleteDescr*>(prev->data);
	next_gt = (prevd->gt) ? prevd : prevd->next_gt;
	if (lt != NULL) {
	    for (;;) {
		prevd->next_lt = this;
		if (prevd->lt != NULL) break;
		prev = prev->prev();
		if (prev == NULL) break;
		prevd = static_cast<SICompleteDescr*>(prev->data);
	    }
	}
    }

    if (next == NULL) {
	next_lt = NULL;
    } else {
	SICompleteDescr * nextd;
	nextd = static_cast<SICompleteDescr*>(next->data);
	next_lt = (nextd->lt) ? nextd: nextd->next_lt;
	if (gt != NULL) {
	    for (;;) {
		nextd->next_gt = this;
		if (nextd->gt != NULL) break;
		next = next->next();
		if (next == NULL) break;
		nextd = static_cast<SICompleteDescr*>(next->data);
	    }
	}
    }
}

bool StringIndex::match(const string & v, CProcessor & p) const { 
    const TSTNode * np = NULL;

    const char * vp = v.c_str();
    const char * s = vp;
    PartialResult res;
    set<Constraint *> ss_cset;

    while ((res = find_partial(&np, &s)) == Found_Partial) {
	//
	// it should always be np->data != NULL
	//
	SIPartialDescr * d;
	d = static_cast<SIPartialDescr*>(np->data);
	if (d->pf != NULL)
	    if (p.process_constraint(d->pf))
		return true;
	if (d->ss != NULL)
	    if (ss_cset.insert(d->ss).second)
		if (p.process_constraint(d->ss))
		    return true;
    }
    if (res == Found_Complete) {
	//
	// it should always be np->data != NULL
	//
	SICompleteDescr * d;
	d = static_cast<SICompleteDescr*>(np->data);
	if (d->eq != NULL)
	    if (p.process_constraint(d->eq))
		return true;
	if (d->sf != NULL)
	    if (p.process_constraint(d->sf))
		return true;
	SICompleteDescr * curs;
	for(curs = d->next_lt; curs != NULL; curs = curs->next_lt) 
	    if (p.process_constraint(curs->lt))
		return true;
	for(curs = d->next_gt; curs != NULL; curs = curs->next_gt) 
	    if (p.process_constraint(curs->gt))
		return true;
    } else if (np != NULL) { 
	//
	// must be Not_Found;
	//
	const TSTNode * curs;
	SICompleteDescr * d;
	if (curs = np->upper_bound(*s)) {
	    d = static_cast<SICompleteDescr*>(curs->data);
	    if (d->lt != NULL) 
		if (p.process_constraint(d->lt))
		    return true;
	    while(d = d->next_lt)
		if (p.process_constraint(d->lt))
		    return true;
	}
	if (curs = np->lower_bound(*s)) {
	    d = static_cast<SICompleteDescr*>(curs->data);
	    if (d->gt != NULL) 
		if (p.process_constraint(d->gt))
		    return true;
	    while(d = d->next_gt)
		if (p.process_constraint(d->gt))
		    return true;
	}
    }

    while(*(s = ++vp) != '\0') {
	np = NULL;
	while ((res = find_partial(&np, &s)) == Found_Partial) {
	    SIPartialDescr * d;
	    d = static_cast<SIPartialDescr*>(np->data);
	    if (d->ss != NULL)
		if (ss_cset.insert(d->ss).second)
		    if (p.process_constraint(d->ss))
			return true;
	}
	if (res == Found_Complete) {
	    SICompleteDescr * d;
	    d = static_cast<SICompleteDescr*>(np->data);
	    if (d->sf != NULL) 
		if (p.process_constraint(d->sf))
		    return true;
	}
    }
    return false;
}
