// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: tst.h,v 1.15 2002/09/30 04:51:49 carzanig Exp $
//
#ifndef TST_H
#define TST_H

#include "ft_allocator.h"
//
// the basic TST algorithms are taken more or less directly from
// R. Sedgewick's "Algorithms in C" 3rd Ed. pp 638--639.  All the rest
// (next, prev, partial insert and match, etc.) is mine.
//
class TST;
class PTST;
class TSTNode;

/** a node in a TST **/
class TSTNode {
public:
    void *		data;

			TSTNode(char c, TSTNode * u);

    TSTNode *		next() const;
    TSTNode *		prev() const;

    TSTNode *		upper_bound(char s) const;
    TSTNode *		lower_bound(char s) const;

    void		print(int level = 0) const;
    bool		is_eos()					const;

private:
    char		c;
    TSTNode *		up;
    TSTNode *		left;
    TSTNode *		middle;
    TSTNode *		right;

    TSTNode *		backtrack_prev() const;
    TSTNode *		backtrack_next() const;

    friend class TST;
    friend class PTST;
};

/** a classic ternary search tree for strings.
 *
 *  This data structure is taken more or less directly from
 *  R. Sedgewick's "Algorithms in C" 3rd Ed. pp 638--639.
 **/
class TST {
public:
			TST();

    TSTNode *		insert(const char * s, FTAllocator & a);
    const TSTNode *	find(const char * s)				const;

    TSTNode *		first()						const;
    TSTNode *		last()						const;
    void		print()						const;
    void		clear();

protected:
    TSTNode *		root;
};

enum PartialResult {
    Found_Partial,
    Found_Complete,
    Not_Found
};

/** this is a TST that allows partial (i.e., prefix) matches and
 *  insertions
 **/
class PTST : public TST {
public:
    PartialResult	find_partial(const TSTNode ** pp, 
				     const char ** s) const;
    TSTNode *		insert_partial(const char * s, FTAllocator & a);
};

#ifdef HAVE_INLINE
#include "tst.icc"
#endif

#endif
